# Vulture allowlist for legitimate unused code
# These are typically type annotations, Protocol imports, and public API methods

# Type annotations and Protocol imports used in string form
Axes
Figure
DifferenceEquationsProtocol
RustModelProtocol
CalibrationParameterTypeProtocol
CalibrationResultWithHistoryProtocol
LossConfigProtocol
OptimizationConfigProtocol
EnsembleSelectionResultProtocol
